;; -*- mode: scheme; coding: utf-8 -*-
;; SPDX-License-Identifier: MIT
;; Copyright © 2020 Göran Weinholt
#!r6rs

;;; SRFI-106 (Socket) interface layer

;; This code is based on the reference implementation at
;; srfi.schemers.org.

;; Limitations:

;; No support for /etc/hosts yet
;; No support for DNS
;; No support for ai-flags
;; No support for the GNU libc name service switch

;; Support for nsswitch would likely require forking off a helper.

;; Without nsswitch, support for DNS will require a bit of code from
;; Industria. A full resolver is not required, only a stub. The stub
;; should look for servers in /etc/resolv.conf. Maybe if dbus is not
;; overcomplicated then it could also support the stub resolver in
;; systemd-networkd, which should hopefully handle the nsswitch stuff.

(library (srfi :106 compat)
  (export
    make-client-socket
    make-server-socket
    socket?
    call-with-socket

    *ai-passive*
    *ai-canonname*
    *ai-numerichost*
    *ai-v4mapped*
    *ai-all*
    *ai-addrconfig*

    (rename (fxior socket-merge-flags))
    socket-purge-flags

    socket-accept
    socket-send socket-send*
    socket-recv socket-recv!
    socket-shutdown
    socket-close
    (rename (AF_UNSPEC *af-unspec*)
            (AF_INET   *af-inet*)
            (AF_INET6  *af-inet6*))
    (rename (SOCK_STREAM *sock-stream*)
            (SOCK_DGRAM  *sock-dgram*))
    (rename (IPPROTO_IP  *ipproto-ip*)
            (IPPROTO_TCP *ipproto-tcp*)
            (IPPROTO_UDP *ipproto-udp*))
    (rename (MSG_PEEK    *msg-peek*)
            (MSG_OOB     *msg-oob*)
            (MSG_WAITALL *msg-waitall*))
    (rename (SHUT_RD   *shut-rd*)
            (SHUT_WR   *shut-wr*)
            (SHUT_RDWR *shut-rdwr*)))
  (import
    (rnrs (6))
    (ip-address)
    (struct pack)
    (loko system fibers)
    (only (loko system unsafe) bytevector-address)
    (loko arch amd64 linux-numbers)
    (loko arch amd64 linux-syscalls))

(define NULL 0)

;; These are from libc, which we don't have access to. We'll need to
;; figure things out ourselves.
(define *ai-passive*     #x0001)
(define *ai-canonname*   #x0002)
(define *ai-numerichost* #x0004)
(define *ai-v4mapped*    #x0008)
(define *ai-all*         #x0010)
(define *ai-addrconfig*  #x0020)

(define-record-type socket
  (sealed #t)
  (fields fd address))

(define-syntax define-optional
  (lambda (x)
    (define (opt-clauses name args* opt*)
      (syntax-case opt* ()
        [() '()]
        [((lhs rhs) (lhs* rhs*) ...)
         (with-syntax ([(args ...) args*])
           #`([(args ...) (#,name args ... rhs)]
              #,@(opt-clauses name #'(args ... lhs) #'((lhs* rhs*) ...))))]))
    (syntax-case x ()
      [(_ (name args ... [(lhs* rhs*) ...])
          . body)
       #`(define name
           (case-lambda
             #,@(opt-clauses #'name #'(args ...) #'((lhs* rhs*) ...))
             [(args ... lhs* ...) . body]))])))

(define (string-split str c)
  (let lp ((start 0) (end 0))
    (cond ((fx=? end (string-length str))
           (list (substring str start end)))
          ((char=? c (string-ref str end))
           (cons (substring str start end)
                 (lp (fx+ end 1) (fx+ end 1))))
          (else
           (lp start (fx+ end 1))))))

(define (socket-purge-flags x . y)
  (apply fxand x (map fxnot y)))

(define-record-type servent
  (sealed #t)
  (fields name port proto aliases))

;; Parse a line in /etc/services
(define (get-service-entry p)
  (let next-line ()
    (let ((line (get-line p)))
      (if (eof-object? line)
          line
          (let ((p (open-string-input-port line)))
            (let lp ((ret '()) (c* '()))
              (let ((c (get-char p)))
                (cond ((or (eof-object? c) (eqv? c #\#) (char-whitespace? c))
                       (let ((ret (if (null? c*)
                                      ret
                                      (cons (list->string (reverse c*)) ret))))
                         (if (or (eof-object? c) (eqv? c #\#))
                             (let ((parts (reverse ret)))
                               (if (not (and (pair? parts) (pair? (cdr parts))))
                                   (next-line)
                                   (let* ((port/protocol (string-split (cadr parts) #\/))
                                          (port (string->number (car port/protocol)))
                                          (protocol (and (pair? (cdr port/protocol))
                                                         (string->symbol (cadr port/protocol)))))
                                     (if (not (and port protocol))
                                         (next-line p)
                                         (make-servent (car parts) port protocol (cddr parts))))))
                             (lp ret '()))))
                      (else
                       (lp ret (cons c c*)))))))))))

;; TODO: GNU/Linux systems use the name service switch, but that
;; requires libc. We can fork off a helper program if that becomes
;; important.
(define service->port
  (let ((cache (make-hashtable equal-hash equal?))) ;("ssh" . tcp) => 22
    (lambda (who service proto)
      (define (fail)
        (error who "Unknown service" service))
      (cond ((string->number service 10) =>
             (lambda (port)
               (if (fx<=? 0 port 65535)
                   port
                   (fail))))
            ((hashtable-ref cache (cons service proto) #f)
             => (lambda (port)
                  (if (eq? port 'no-entry)
                      (fail)
                      port)))
            ((not (file-exists? "/etc/services"))
             (fail))
            (else
             (call-with-input-file "/etc/services"
               (lambda (p)
                 (let lp ()
                   (let ((entry (get-service-entry p)))
                     (cond ((eof-object? entry)
                            (hashtable-set! cache (cons service proto) 'no-entry)
                            (fail))
                           ((and (or (not proto)
                                     (eq? (servent-proto entry) proto))
                                 (or (string=? (servent-name entry) service)
                                     (member service (servent-aliases entry))))
                            (hashtable-set! cache (cons service proto) (servent-port entry))
                            (servent-port entry))
                           (else (lp))))))))))))

(define (sockaddr-family buf)
  (unpack "S" buf))

(define (sockaddr->string buf)
  (define (sockaddr_in6->string buf)
    (let ((ip (make-bytevector 16)))
      (bytevector-copy! buf offsetof-sockaddr_in6-sin6_addr ip 0 16)
      (let ((port (unpack "!S" buf offsetof-sockaddr_in6-sin6_port)))
        (string-append "[" (ipv6->string ip) "]:" (number->string port)))))
  (define (sockaddr_in->string buf)
    (let ((ip (make-bytevector 4)))
      (pack! "L" ip 0 (unpack "L" buf offsetof-sockaddr_in-sin_addr))
      (let ((port (unpack "!S" buf offsetof-sockaddr_in-sin_port)))
        (string-append (ipv4->string ip) ":" (number->string port)))))
  (let ((family (sockaddr-family buf)))
    (cond ((eqv? family AF_INET)
           (sockaddr_in->string buf))
          ((eqv? family AF_INET6)
           (sockaddr_in6->string buf))
          (else
           (call-with-string-output-port
             (lambda (p)
               (display "unknown family " p)
               (display buf p)))))))

(define (filename->sockaddr_un who fn)
  (unless (string? fn)
    (assertion-violation who "Expected a string" fn))
  (string-for-each
   (lambda (c)
     (when (eqv? c #\nul)
       (raise (condition
               (make-who-condition who)
               (make-i/o-filename-error fn)
               (make-message-condition "Unrepresentable filename")
               (make-irritants-condition (list fn))))))
   fn)
  (let ((addr (make-bytevector sizeof-sockaddr_un 0))
        (filename-bv (string->utf8 fn)))
    (bytevector-u16-native-set! addr offsetof-sockaddr_un-sun_family AF_LOCAL)
    ;; XXX: Linux can handle a missing NUL character at the very end
    ;; of the buffer
    (unless (fx<? (bytevector-length filename-bv)
                  (fx- sizeof-sockaddr_un offsetof-sockaddr_un-sun_path))
      (assertion-violation who "The filename is too long" fn))
    (bytevector-copy! filename-bv 0
                      addr offsetof-sockaddr_un-sun_path
                      (bytevector-length filename-bv))
    addr))

(define (make-sockaddr who node service ai-family proto)
  (cond
    ((eqv? ai-family AF_LOCAL)
     (filename->sockaddr_un who node))
    ((or (eqv? ai-family AF_INET) (eqv? ai-family AF_INET6))
     (let ((port (service->port who service proto))
           (node/scope (string-split node #\%)))
       (let ((node (car node/scope))
             (scope (if (pair? (cdr node/scope)) (cadr node/scope) #f)))
         (cond ((string->ipv6 node) =>
                (lambda (ip)
                  (let ((addr (make-bytevector sizeof-sockaddr_in6)))
                    (bytevector-u16-native-set! addr offsetof-sockaddr_in6-sin6_family AF_INET6)
                    (bytevector-u16-set! addr offsetof-sockaddr_in6-sin6_port port (endianness big))
                    (bytevector-copy! ip 0 addr offsetof-sockaddr_in6-sin6_addr (bytevector-length ip))
                    ;; TODO: Look up the scope id
                    (bytevector-u32-native-set! addr offsetof-sockaddr_in6-sin6_scope_id 0)
                    addr)))
               ((string->ipv4 node) =>
                (lambda (ip)
                  (let ((addr (make-bytevector sizeof-sockaddr_in)))
                    (bytevector-u16-native-set! addr offsetof-sockaddr_in-sin_family AF_INET)
                    (bytevector-u16-set! addr offsetof-sockaddr_in-sin_port port (endianness big))
                    (bytevector-copy! ip 0 addr offsetof-sockaddr_in-sin_addr (bytevector-length ip))
                    addr)))
               ((equal? node "localhost")
                ;; Cheating, cheating, cheating.
                (make-sockaddr who "127.0.0.1" service ai-family proto))
               (else
                (assertion-violation who "TODO: DNS and hosts support" node service))))))
    (else
     (assertion-violation who "Unknown address family" node service))))

(define (accept fd addr addrlen flags)
  (sys_accept4 fd (bytevector-address addr) (bytevector-address addrlen) flags
               (lambda (errno)
                 (cond ((eqv? errno EAGAIN)
                        (wait-for-readable fd)
                        (accept fd addr addrlen flags))
                       ((eqv? errno EINTR)
                        (accept fd addr addrlen flags))
                       (else
                        (raise (condition
                                (make-syscall-error 'accept4 errno)
                                (make-irritants-condition
                                 (list fd addr addrlen flags)))))))))

(define (setsockopt fd level optname optval)
  (sys_setsockopt fd level optname (bytevector-address optval) (bytevector-length optval)))

(define (ai-junk->protocol ai-family ai-socktype ai-protocol)
  (cond ((eqv? ai-protocol IPPROTO_UDP) 'udp)
        ((eqv? ai-protocol IPPROTO_TCP) 'tcp)
        ((eqv? ai-protocol IPPROTO_SCTP) 'sctp)
        (else #f)))

(define-optional (make-client-socket node service
                                     [(ai-family AF_INET)
                                      (ai-socktype SOCK_STREAM)
                                      (ai-flags (fxior *ai-v4mapped* *ai-addrconfig*))
                                      (ai-protocol IPPROTO_IP)])
  (let* ((addr (make-sockaddr 'make-client-socket node service ai-family
                              (ai-junk->protocol ai-family ai-socktype ai-protocol)))
         (fd (sys_socket (sockaddr-family addr)
                         (fxior ai-socktype SOCK_CLOEXEC SOCK_NONBLOCK)
                         ai-protocol)))
    (let retry ()
      (sys_connect fd (bytevector-address addr) (bytevector-length addr)
                   (lambda (errno)
                     (cond ((eqv? errno EINTR)
                            (retry))
                           ((eqv? errno EINPROGRESS)
                            (wait-for-writable fd)
                            (let* ((statusbv (make-bytevector sizeof-int))
                                   (statuslen (if (eqv? sizeof-void* 8)
                                                  (pack "Q" sizeof-int)
                                                  (pack "L" sizeof-int))))
                              (sys_getsockopt fd SOL_SOCKET SO_ERROR
                                              (bytevector-address statusbv)
                                              (bytevector-address statuslen))
                              (let ((status (if (eqv? sizeof-int 8)
                                                (unpack "Q" statusbv)
                                                (unpack "L" statusbv))))
                                (if (eqv? status 0)
                                    0
                                    (raise (condition
                                            (make-who-condition 'make-client-socket)
                                            (make-syscall-error 'connect status)
                                            (make-irritants-condition (list node service))))))))
                           ((eqv? errno EAGAIN)
                            ;; This is for local sockets.
                            ;; FIXME: Manpages are unclear.
                            (wait-for-writable fd)
                            (retry))
                           (else
                            (sys_close fd)
                            (raise (condition
                                    (make-who-condition 'make-client-socket)
                                    (make-syscall-error 'connect errno)
                                    (make-irritants-condition (list fd addr)))))))))
    (make-socket fd (sockaddr->string addr))))

(define-optional (make-server-socket service
                                     [(ai-family AF_INET)
                                      (ai-socktype SOCK_STREAM)
                                      (ai-protocol IPPROTO_IP)])
  (let ((addr (make-sockaddr 'make-server-socket
                             (if (eqv? ai-family AF_INET)
                                 "0.0.0.0"
                                 "::1")
                             service ai-family
                             (ai-junk->protocol ai-family ai-socktype ai-protocol))))
    (let ((fd (sys_socket ai-family
                          (fxior ai-socktype SOCK_CLOEXEC SOCK_NONBLOCK)
                          ai-protocol)))
      (let ((optval (pack "l" 1)))
        (sys_setsockopt fd SOL_SOCKET SO_REUSEADDR
                        (bytevector-address optval) (bytevector-length optval)))
      (sys_bind fd (bytevector-address addr) (bytevector-length addr))
      (sys_listen fd 1000)
      (make-socket fd (sockaddr->string addr)))))

(define (call-with-socket socket proc)
  (assert (socket? socket))
  (let-values ([v (proc socket)])
    (socket-close socket)
    (apply values v)))

(define (socket-accept socket)
  (let ((addr (make-bytevector sizeof-sockaddr_in6))
        (addrlen (pack "q" sizeof-sockaddr_in6))
        (flags (bitwise-ior SOCK_NONBLOCK SOCK_CLOEXEC))
        (sfd (socket-fd socket)))
    (let ((fd (let retry ()
                (sys_accept4 sfd (bytevector-address addr)
                             (bytevector-address addrlen) flags
                             (lambda (errno)
                               (cond ((eqv? errno EAGAIN)
                                      (wait-for-readable sfd)
                                      (retry))
                                     ((eqv? errno EINTR)
                                      (retry))
                                     (else
                                      (raise (condition
                                              (make-who-condition 'socket-accept)
                                              (make-syscall-error 'accept4 errno)
                                              (make-irritants-condition (list socket)))))))))))
      (make-socket fd (sockaddr->string addr)))))

(define-optional (socket-send socket buf [(flags 0)])
  (sys_sendto (socket-fd socket) (bytevector-address buf) (bytevector-length buf)
              flags NULL 0
              (lambda (errno)
                (cond ((eqv? errno EAGAIN)
                       (wait-for-writable (socket-fd socket))
                       (socket-send socket buf flags))
                      ((eqv? errno EINTR)
                       (socket-send socket buf flags))
                      (else
                       (raise (condition
                               (make-who-condition 'socket-send)
                               (make-syscall-error 'sendto errno)
                               (make-irritants-condition
                                (list (socket-fd socket))))))))))

(define (socket-send* socket buf start count) ;not standard
  (define flags 0)
  (assert (fx<=? 0 start (fx+ start count) (bytevector-length buf)))
  (sys_sendto (socket-fd socket) (fx+ (bytevector-address buf) start) count
              flags NULL 0
              (lambda (errno)
                (cond ((eqv? errno EAGAIN)
                       (wait-for-writable (socket-fd socket))
                       (socket-send socket buf flags))
                      ((eqv? errno EINTR)
                       (socket-send socket buf flags))
                      (else
                       (raise (condition
                               (make-who-condition 'socket-send)
                               (make-syscall-error 'sendto errno)
                               (make-irritants-condition
                                (list (socket-fd socket))))))))))

(define-optional (socket-recv socket size [(flags 0)])
  (let ((buf (make-bytevector size)))
    (let ((len (let retry ()
                 (sys_recvfrom (socket-fd socket) (bytevector-address buf)
                               (bytevector-length buf)
                               flags NULL 0
                               (lambda (errno)
                                 (cond ((eqv? errno EAGAIN)
                                        (wait-for-readable (socket-fd socket))
                                        (retry))
                                       ((eqv? errno EINTR)
                                        (retry))
                                       (else
                                        (raise (condition
                                                (make-who-condition 'socket-recv)
                                                (make-syscall-error 'recvfrom errno)
                                                (make-irritants-condition
                                                 (list (socket-fd socket))))))))))))
      ;; TODO: Truncate buf in-place instead
      (if (fx<? len size)
          (let ((ret (make-bytevector len)))
            (bytevector-copy! buf 0 ret 0 len)
            ret)
          buf))))

(define (socket-recv! socket buf start size) ;not standard
  (define flags 0)
  (assert (fx<=? 0 start (fx+ start size) (bytevector-length buf)))
  (let lp ()
    (sys_recvfrom (socket-fd socket) (fx+ start (bytevector-address buf))
                  size flags NULL 0
                  (lambda (errno)
                    (cond ((eqv? errno EAGAIN)
                           (wait-for-readable (socket-fd socket))
                           (lp))
                          ((eqv? errno EINTR)
                           (lp))
                          (else
                           (raise (condition
                                   (make-who-condition 'socket-recv)
                                   (make-syscall-error 'recvfrom errno)
                                   (make-irritants-condition
                                    (list (socket-fd socket)))))))))))

(define (socket-shutdown socket how)
  (sys_shutdown (socket-fd socket) how))

(define (socket-close socket)
  (sys_close (socket-fd socket))))
